__author__ = 'Robbert Harms'
__date__ = '2020-01-25'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

import pyopencl as cl


class Processor:

    def process(self, is_blocking=False, wait_for=None):
        """Enqueue all compute kernels for this processor.

        This may enqueue multiple kernels to multiple devices.

        Args:
            wait_for (Dict[CLEnvironment: cl.Event]): mapping CLEnvironments to events we should wait
                on before enqueuing work. This will wait on all events sharing the same context.
            is_blocking (boolean): if the call to the processor is a blocking call or not

        Returns:
            Dict[CLEnvironment: cl.Event]: events generated by this processor
        """
        raise NotImplementedError()

    def flush(self):
        """Enqueues a flush operation to all the queues."""
        raise NotImplementedError()

    def finish(self):
        """Enqueues a finish operation to all the queues."""
        raise NotImplementedError()


class MultiDeviceProcessor(Processor):

    def __init__(self, kernels, kernel_data, cl_environments, load_balancer,
                 nmr_instances, use_local_reduction=False, local_size=None, do_data_transfers=True):
        """Create a processor for the given function and inputs.

        Args:
            kernels (dict): for each CL environment the kernel to use
            kernel_data (dict): the input data for the kernels
            cl_environments (List[mot.lib.cl_environments.CLEnvironment]): the list of CL environment to use
                for executing the kernel
            load_balancer (mot.lib.load_balancers.LoadBalancer): the load balancer to use
            nmr_instances (int): the number of parallel processes to run.
            use_local_reduction (boolean): set this to True if you want to use local memory reduction in
                 evaluating this function. If this is set to True we will multiply the global size
                 (given by the nmr_instances) by the work group sizes.
            local_size (int): can be used to specify the exact local size (workgroup size) the kernel must use.
            do_data_transfers (boolean): if we should do data transfers from host to device and back for evaluating
                this function. For better control set this to False and use the method
                ``enqueue_device_access()`` and ``enqueue_host_access`` of the KernelData to set the data.
        """
        self._subprocessors = []
        self._do_data_transfers = do_data_transfers
        self._kernel_data = kernel_data
        self._cl_environments = cl_environments

        batches = load_balancer.get_division(cl_environments, nmr_instances)
        for ind, cl_environment in enumerate(cl_environments):
            kernel = kernels[cl_environment]

            if use_local_reduction:
                if local_size:
                    workgroup_size = local_size
                else:
                    workgroup_size = kernel.get_work_group_info(
                        cl.kernel_work_group_info.PREFERRED_WORK_GROUP_SIZE_MULTIPLE, cl_environment.device)
            else:
                workgroup_size = 1

            batch_start, batch_end = batches[ind]
            if batch_end - batch_start > 0:
                processor = ProcessKernel(kernel, kernel_data.values(), cl_environment,
                                          batch_end - batch_start, workgroup_size,
                                          instance_offset=batch_start)
                self._subprocessors.append(processor)

    def process(self, is_blocking=False, wait_for=None):
        if self._do_data_transfers:
            for ind, kernel_data in enumerate(self._kernel_data.values()):
                wait_for = kernel_data.enqueue_device_access(self._cl_environments, is_blocking=False,
                                                             wait_for=wait_for)

        events = {}
        for worker in self._subprocessors:
            events.update(worker.process(wait_for=wait_for))
            worker.flush()

        if self._do_data_transfers:
            for ind, kernel_data in enumerate(self._kernel_data.values()):
                events = kernel_data.enqueue_host_access(self._cl_environments, is_blocking=False, wait_for=events)

        return events

    def flush(self):
        for worker in self._subprocessors:
            worker.flush()

    def finish(self):
        for worker in self._subprocessors:
            worker.finish()


class ProcessKernel(Processor):

    def __init__(self, kernel, kernel_data, cl_environment, global_nmr_instances, workgroup_size, instance_offset=None):
        """Simple processor which can execute the provided (compiled) kernel with the provided data.

        Args:
            kernel: a pyopencl compiled kernel program
            kernel_data (List[mot.lib.utils.KernelData]): the kernel data to load as input to the kernel
            cl_environment (mot.lib.cl_environments.CLEnvironment): the CL environment to use for executing the kernel
            global_nmr_instances (int): the global work size, this will internally be multiplied by the
                local workgroup size.
            workgroup_size (int): the local size (workgroup size) the kernel must use
            instance_offset (int): the offset for the global id, this will be multiplied with the local workgroup size.
        """
        self._kernel = kernel
        self._kernel_data = kernel_data
        self._cl_environment = cl_environment
        self._global_nmr_instances = global_nmr_instances
        self._instance_offset = instance_offset or 0
        self._kernel.set_scalar_arg_dtypes(self._flatten_list([d.get_scalar_arg_dtypes() for d in self._kernel_data]))
        self._workgroup_size = workgroup_size

    def process(self, is_blocking=False, wait_for=None):
        wait_for = wait_for or {}
        if self._cl_environment in wait_for:
            wait_for = [wait_for[self._cl_environment]]
        else:
            wait_for = None

        event = self._kernel(
            self._cl_environment.queue,
            (int(self._global_nmr_instances * self._workgroup_size),),
            (int(self._workgroup_size),),
            *self._flatten_list([data.get_kernel_inputs(self._cl_environment, self._workgroup_size)
                                 for data in self._kernel_data]),
            global_offset=(int(self._instance_offset * self._workgroup_size),),
            wait_for=wait_for)

        if is_blocking:
            event.wait()

        return {self._cl_environment: event}

    def flush(self):
        self._cl_environment.queue.flush()

    def finish(self):
        self._cl_environment.queue.finish()

    def _flatten_list(self, l):
        return_l = []
        for e in l:
            return_l.extend(e)
        return return_l


class DeviceAccess(Processor):

    def __init__(self, kernel_data, cl_environments):
        """A processor to enqueue device access for all the provided kernel data.

        Args:
            kernel_data (List[mot.lib.utils.KernelData]): the input data for the kernels
            cl_environments (List[mot.lib.cl_environments.CLEnvironment]): the list of CL environment to use
                for executing the kernel
        """
        self._kernel_data = kernel_data
        self._cl_environments = cl_environments

    def process(self, is_blocking=False, wait_for=None):
        events = None
        for ind, kernel_data in enumerate(self._kernel_data):
            events = kernel_data.enqueue_device_access(self._cl_environments, is_blocking=False, wait_for=wait_for)
        return events

    def flush(self):
        for env in self._cl_environments:
            env.queue.flush()

    def finish(self):
        for env in self._cl_environments:
            env.queue.finish()


class HostAccess(Processor):

    def __init__(self, kernel_data, cl_environments):
        """A processor to enqueue device access for all the provided kernel data.

        Args:
            kernel_data (List[mot.lib.utils.KernelData]): the input data for the kernels
            cl_environments (List[mot.lib.cl_environments.CLEnvironment]): the list of CL environment to use
                for executing the kernel
        """
        self._kernel_data = kernel_data
        self._cl_environments = cl_environments

    def process(self, is_blocking=False, wait_for=None):
        events = None
        for ind, kernel_data in enumerate(self._kernel_data):
            events = kernel_data.enqueue_host_access(self._cl_environments, is_blocking=False, wait_for=wait_for)
        return events

    def flush(self):
        for env in self._cl_environments:
            env.queue.flush()

    def finish(self):
        for env in self._cl_environments:
            env.queue.finish()



from mot.library_functions.base import SimpleCLLibrary

__author__ = 'Robbert Harms'
__date__ = '2018-05-07'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'

from mot.library_functions.continuous_distributions.gamma import igamc, igamci


class invgamma_pdf(SimpleCLLibrary):
    def __init__(self):
        r"""Computes the inverse-Gamma probability density function using the shape and scale parameterization.

        This computes the inverse-Gamma PDF as:

        .. math::

            f(x; \alpha, \beta) = \frac{\beta^\alpha}{\Gamma(\alpha)} (1/x)^{\alpha + 1}\exp\left(-\beta/x\right)

        With :math:`x` the desired position, :math:`\alpha` the shape and :math:`\beta` the scale.
        """
        super().__init__('''
            double invgamma_pdf(double x, double shape, double scale){
                return exp(invgamma_logpdf(x, shape, scale));
            }
        ''', dependencies=[invgamma_logpdf()])


class invgamma_logpdf(SimpleCLLibrary):
    def __init__(self):
        r"""Computes the log of the inverse-Gamma probability density function with shape and scale parameterization.
        """
        super().__init__('''
            double invgamma_logpdf(double x, double shape, double scale){
                double _x = x / scale;
                return -(shape+1) * log(_x) - lgamma(shape) - 1.0/_x - log(scale);
            }
        ''')


class invgamma_cdf(SimpleCLLibrary):
    def __init__(self):
        r"""Calculate the Cumulative Distribution Function of the inverse-Gamma function.

        This implementation is copied from SciPy (09-01-2020).

        Function arguments:

         * shape: the shape parameter of the gamma distribution (often denoted :math:`\alpha`)
         * scale: the scale parameter of the gamma distribution (often denoted :math:`\beta`)
        """
        super().__init__('''
            double invgamma_cdf(double x, double shape, double scale){
                return igamc(shape, scale / x);
            }
        ''', dependencies=(igamc(),))


class invgamma_ppf(SimpleCLLibrary):

    def __init__(self):
        """Computes the inverse of the cumulative distribution function of the inverse-Gamma distribution.

        This is the inverse of the inverse-Gamma CDF.
        """
        super().__init__('''
            double invgamma_ppf(double y, double shape, double scale){
                return scale / igamci(shape, y);
            }
        ''', dependencies=(igamci(),))



"""Contains the runtime configuration of MOT.

This consists of two parts, functions to get the current runtime settings and configuration actions to update these
settings. To set a new configuration, create a new :py:class:`ConfigAction` and use this within a context environment
using :py:func:`config_context`. Example:

.. code-block:: python

    from mot.configuration import RuntimeConfigurationAction, config_context

    with config_context(RuntimeConfigurationAction(...)):
        ...

"""
import collections
from contextlib import contextmanager
import numpy as np

from mot.lib.load_balancers import EvenDistribution, FractionalLoad
from .lib.cl_environments import CLEnvironmentFactory

__author__ = 'Robbert Harms'
__date__ = "2015-07-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert@xkls.nl"


"""The runtime configuration, this can be overwritten at run time.

For any of the AbstractCLRoutines it holds that if no suitable defaults are given we use the ones provided by this
module. This entire module acts as a singleton containing the current runtime configuration.
"""
_config = {
    'cl_environments': CLEnvironmentFactory.smart_device_selection(preferred_device_type='GPU'),
    'compile_flags': ['-cl-denorms-are-zero', '-cl-mad-enable', '-cl-no-signed-zeros'],
    'double_precision': False,
    'load_balancer': EvenDistribution()
}


def get_cl_environments():
    """Get the current CL environment to use during CL calculations.

    Returns:
        list of CLEnvironment: the current list of CL environments.
    """
    return _config['cl_environments']


def set_cl_environments(cl_environments):
    """Set the current CL environments to the given list

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        cl_environments (List[Union[CLEnvironment, int]]): the new list of CL environments. Can also be a list
            of integers (or a single integer) indicing into :func:`CLEnvironmentFactory.smart_device_selection`.

    Raises:
        ValueError: if the list of environments is empty
    """
    if not isinstance(cl_environments, collections.Iterable):
        cl_environments = [cl_environments]

    final_environments = []
    all_environments = CLEnvironmentFactory.smart_device_selection()

    for environment in cl_environments:
        if isinstance(environment, int):
            final_environments.append(all_environments[environment])
        else:
            final_environments.append(environment)

    if not final_environments:
        raise ValueError('The list of CL Environments is empty.')

    _config['cl_environments'] = final_environments


def get_compile_flags():
    """Get the default compile flags to use in a CL routine.

    Returns:
        list: the default list of compile flags we wish to use
    """
    return list(_config['compile_flags'])


def set_compile_flags(compile_flags):
    """Set the current compile flags.

    Args:
        compile_flags (list): the new list of compile flags
    """
    _config['compile_flags'] = compile_flags


def set_default_proposal_update(proposal_update):
    """Set the default proposal update function to use in sample.

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        mot.model_building.parameter_functions.proposal_updates.ProposalUpdate: the new proposal update function
            to use by default if no specific one is provided.
    """
    _config['default_proposal_update'] = proposal_update


def use_double_precision():
    """Check if we run the computations in default precision or not.

    Returns:
        boolean: if we run the computations in double precision or not
    """
    return _config['double_precision']


def set_use_double_precision(double_precision):
    """Set the default use of double precision.

    Returns:
        boolean: if we use double precision by default or not
    """
    _config['double_precision'] = double_precision


def get_load_balancer():
    """Get the current load balancing settings.

    The load balancing determines how much work each of the CLEnvironments (i.e. devices) will perform.

    Returns:
        mot.lib.load_balancers.LoadBalancer: the current load balancing strategy
    """
    return _config['load_balancer']


def set_load_balancer(load_balancer):
    """Set the default load balancer.

    Args:
        mot.lib.load_balancers.LoadBalancer: the new load balancing strategy
    """
    _config['load_balancer'] = load_balancer


@contextmanager
def config_context(config_action):
    """Creates a context in which the config action is applied and unapplies the configuration after execution.

    Args:
        config_action (ConfigAction): the configuration action to use
    """
    config_action.apply()
    yield
    config_action.unapply()


class ConfigAction:

    def __init__(self):
        """Defines a configuration action for use in a configuration context.

        This should define an apply and unapply function that sets and unsets the configuration options.

        The applying action needs to remember the state before the application of the action.
        """

    def apply(self):
        """Apply the current action to the current runtime configuration."""

    def unapply(self):
        """Reset the current configuration to the previous state."""


class SimpleConfigAction(ConfigAction):

    def __init__(self):
        """Defines a default implementation of a configuration action.

        This simple config implements a default ``apply()`` method that saves the current state and a default
        ``unapply()`` that restores the previous state.

        For developers, it is easiest to implement ``_apply()`` such that you do not manually need to store the old
        configuraration.
        """
        super().__init__()
        self._old_config = {}

    def apply(self):
        """Apply the current action to the current runtime configuration."""
        self._old_config = {k: v for k, v in _config.items()}
        self._apply()

    def unapply(self):
        """Reset the current configuration to the previous state."""
        for key, value in self._old_config.items():
            _config[key] = value

    def _apply(self):
        """Implement this function add apply() logic after this class saves the current config."""


class CLRuntimeAction(SimpleConfigAction):

    def __init__(self, cl_runtime_info):
        """Set the current configuration to use the information in the given configuration action.

        Args:
            cl_runtime_info (CLRuntimeInfo): the runtime info with the configuration options
        """
        super().__init__()
        self._cl_runtime_info = cl_runtime_info

    def _apply(self):
        set_cl_environments(self._cl_runtime_info.cl_environments)
        set_compile_flags(self._cl_runtime_info._compile_flags)
        set_use_double_precision(self._cl_runtime_info.double_precision)
        set_load_balancer(self._cl_runtime_info.load_balancer)


class RuntimeConfigurationAction(SimpleConfigAction):

    def __init__(self, cl_environments=None, compile_flags=None, double_precision=None, load_balancer=None):
        """Updates the runtime settings.

        Args:
            cl_environments (list of CLEnvironment): the new CL environments we wish to use for future computations
            compile_flags (list): the list of compile flags to use during analysis.
            double_precision (boolean): if we compute in double precision or not
            load_balancer (mot.lib.load_balancers.LoadBalancer): the new load balancing strategy
        """
        super().__init__()
        self._cl_environments = cl_environments
        self._compile_flags = compile_flags
        self._double_precision = double_precision
        self._load_balancer = load_balancer

    def _apply(self):
        if self._cl_environments is not None:
            set_cl_environments(self._cl_environments)

        if self._compile_flags is not None:
            set_compile_flags(self._compile_flags)

        if self._double_precision is not None:
            set_use_double_precision(self._double_precision)

        if self._load_balancer is not None:
            set_load_balancer(self._load_balancer)


class VoidConfigurationAction(ConfigAction):

    def __init__(self):
        """Does nothing, useful as a default config action.
        """
        super().__init__()


class CLRuntimeInfo:

    def __init__(self, cl_environments=None, compile_flags=None, double_precision=None, load_balancer=None):
        """All information necessary for applying operations using OpenCL.

        Args:
            cl_environments (List[Union[mot.lib.cl_environments.CLEnvironment, int]]
                             or mot.lib.cl_environments.CLEnvironment or int): The list of CL environments
                used by this routine. If None is given we use the defaults in the current configuration.
                Elements can either be a CLEnvironment object or an integer indexing the devices returned by
                :func:`mot.smart_device_selection`.
            compile_flags (list): the list of compile flags to use during analysis.
            double_precision (boolean): if we apply the computations in double precision or in single float precision.
                By default we go for single float precision.
            load_balancer (mot.lib.load_balancers.LoadBalancer or Tuple[float]): the load balancer to use
                for the computations. Can either be a load balancer or a tuple with fractional loads per device.
        """
        self._cl_environments = self._load_environments(cl_environments)
        self._compile_flags = tuple(compile_flags or get_compile_flags())
        self._double_precision = double_precision
        self._load_balancer = self._prepare_load_balancer(load_balancer)

        if self._double_precision is None:
            self._double_precision = use_double_precision()

    @staticmethod
    def _load_environments(environments):
        """Load the environments from a polymorphic datatype."""
        if environments is None:
            environments = get_cl_environments()

        if not isinstance(environments, collections.Iterable):
            environments = [environments]

        final_environments = []
        all_environments = CLEnvironmentFactory.smart_device_selection()

        for environment in environments:
            if isinstance(environment, int):
                final_environments.append(all_environments[environment])
            else:
                final_environments.append(environment)

        return final_environments

    @staticmethod
    def _prepare_load_balancer(load_balancer):
        """Load the load balancer from a polymorphic datatype."""
        if load_balancer is None:
            return get_load_balancer()
        if isinstance(load_balancer, collections.Iterable):
            return FractionalLoad(load_balancer)
        return load_balancer or EvenDistribution()

    @property
    def cl_environments(self):
        return self._cl_environments

    @property
    def mot_float_dtype(self):
        if self._double_precision:
            return np.float64
        return np.float32

    @property
    def double_precision(self):
        return self._double_precision

    @property
    def compile_flags(self):
        """Get all defined compile flags."""
        return self._compile_flags

    @property
    def load_balancer(self):
        return self._load_balancer



import numpy as np
from mot.lib.utils import split_in_batches

__author__ = 'Robbert Harms'
__date__ = '2019-12-20'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'


class LoadBalancer:
    """Load balancers allow for fine-tuned division of labour over multiple CL devices.

    The idea is that when evaluating a kernel, the work can be split over all selected devices. The typical
    division is an even distribution, but this can be fine-tuned if one device is faster than others.
    """

    def get_division(self, cl_environments, nmr_instances):
        """Get the proposed division of labour for dividing the given number of instances over the given environments.

        Args:
            cl_environments (List[mot.lib.cl_environments.CLEnvironment]): the environments used in the load balancing.
            nmr_instances (int): the number of work instances to divide

        Returns:
            List[Tuple[int, int]]: list with (batch_start, batch_end) tuples dividing the number of instances.
        """
        raise NotImplementedError()


class EvenDistribution(LoadBalancer):
    """Evenly distribute the work over all available environments."""

    def get_division(self, cl_environments, nmr_instances):
        return list(split_in_batches(nmr_instances, nmr_batches=len(cl_environments)))


class FractionalLoad(LoadBalancer):

    def __init__(self, fractions):
        """Balance the load according to a specified fraction per devices.

        This class will round the work items such that all work gets done.

        Args:
            fractions (Tuple[float]): for each device the fraction of work for that device
        """
        fractions = np.array(fractions)
        self._fractions = fractions / np.sum(fractions)

    def get_division(self, cl_environments, nmr_instances):
        if len(cl_environments) != len(self._fractions):
            raise ValueError('The number of devices does not match the number of specified loads.')

        batches = []
        offset = 0
        elements_left = nmr_instances
        for fraction in self._fractions[:-1]:
            batch_size = min(elements_left, int(np.round(fraction * nmr_instances)))
            batches.append((offset, offset + batch_size))
            elements_left -= batch_size
            offset += batch_size

        batches.append((offset, offset + elements_left))
        return batches



from collections import Iterable
from copy import copy
import tatsu
from textwrap import dedent, indent
import pyopencl as cl
from mot.configuration import CLRuntimeInfo
from mot.lib.cl_processors import MultiDeviceProcessor
from mot.lib.kernel_data import Zeros
from mot.lib.utils import split_cl_function, convert_inputs_to_kernel_data, get_cl_utility_definitions

__author__ = 'Robbert Harms'
__date__ = '2017-08-31'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'


class CLCodeObject:
    """Interface for basic code objects."""

    def get_cl_code(self):
        """Get the CL code for this code object and all its dependencies, with include guards.

        Returns:
            str: The CL code for inclusion in a kernel.
        """
        raise NotImplementedError()


class CLFunction(CLCodeObject):
    """Interface for a basic CL function."""

    def is_kernel_func(self):
        """Check if this function is a kernel function or not.

        Kernel functions have the keyword 'kernel' in them, like: ``kernel void foo();``.

        Returns:
            boolean: if this function is a kernel function or not
        """
        raise NotImplementedError()

    def get_kernel_wrapped(self, input_data, nmr_instances, kernel_name=None):
        """Wrap the current CLFunction with a kernel CLFunction.

        This may return self if the current function is already a kernel function.

        The idea is that we can have a function like:

        .. code-block:: c

            int foo(void* data){...}

        That is, without the ``kernel`` modifier. This function can not readily be executed on a CL device.
        To make life easy, this method wraps the current CL code in a kernel like this:

        .. code-block:: c

            int foo(void* data){...}

            kernel void kernel_foo(...){
                ... data = ...;
                int result = foo(&data);
                __return_values[gid] = result
            }

        And then kernel_foo can be executed on the device. Note that if the function we are wrapping has a non-void
        return type, we return an Zeros kernel data element to be added to the kernel data when calling
        the wrapped function.

        In order to generate the correct kernel arguments, this method needs to know which data will be loaded and
        with which signature. This is done by providing it with all the kernel inputs you wish to load into the
        kernel when evaluating the function. The generated kernel function will do automatic data marshalling
        from all the :class:`mot.lib.kernel_data.KernelData` inputs to the inputs for the wrapped function.

        Args:
            input_data (Dict[str: mot.lib.kernel_data.KernelData]): mapping parameter names to kernel data objects.
            nmr_instances (int): the number of instances we run the data with, needed to generate the possible
                additional kernel data.
            kernel_name (str): the name of the generated kernel function. If not given it will be called
                ``kernel_<CLFunction.get_cl_function_name()>``.

        Returns:
            Tuple[CLFunction, Optional[Dict[str, KernelData]]]: A tuple with a CL function with :meth:`is_kernel_func`
                set to True (might return 'self'), and an optional dictionary with a KernelData element
                for storing the possible CL function return values.
        """
        raise NotImplementedError()

    def get_return_type(self):
        """Get the type (in CL naming) of the returned value from this function.

        Returns:
            str: The return type of this CL function. (Examples: double, int, double4, ...)
        """
        raise NotImplementedError()

    def get_cl_function_name(self):
        """Return the calling name of the implemented CL function

        Returns:
            str: The name of this CL function
        """
        raise NotImplementedError()

    def get_parameters(self):
        """Return the list of parameters from this CL function.

        Returns:
            list of :class:`mot.lib.cl_function.CLFunctionParameter`: list of the parameters in this
                model in the same order as in the CL function"""
        raise NotImplementedError()

    def get_signature(self):
        """Get the CL signature of this function.

        Returns:
            str: the CL code for the signature of this CL function.
        """
        raise NotImplementedError()

    def get_cl_code(self):
        """Get the function code for this function and all its dependencies, with include guards.

        Returns:
            str: The CL code for inclusion in a kernel.
        """
        raise NotImplementedError()

    def get_cl_body(self):
        """Get the CL code for the body of this function.

        Returns:
            str: the CL code of this function body
        """
        raise NotImplementedError()

    def evaluate(self, inputs, nmr_instances, use_local_reduction=False, local_size=None, cl_runtime_info=None,
                 do_data_transfers=True, is_blocking=True, return_events=False, wait_for=None):
        """Evaluate this function for each set of given parameters.

        Given a set of input parameters, this model will be evaluated for every parameter set.
        This function will convert possible dots in the parameter names to underscores for use in the CL kernel.

        Args:
            inputs (Iterable[Union(ndarray, mot.lib.utils.KernelData)]
                    or Mapping[str: Union(ndarray, mot.lib.utils.KernelData)]): for each CL function parameter
                the input data. Each of these input datasets must either be a scalar or be of equal length in the
                first dimension. The elements can either be raw ndarrays or KernelData objects.
                If an ndarray is given we will load it read/write by default. You can provide either an iterable
                with one value per parameter, or a mapping with for every parameter a corresponding value.
            nmr_instances (int): the number of parallel processes to run.
            use_local_reduction (boolean): set this to True if you want to use local memory reduction in
                 evaluating this function. If this is set to True we will multiply the global size
                 (given by the nmr_instances) by the work group sizes.
            local_size (int): can be used to specify the exact local size (workgroup size) the kernel must use.
            cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information for execution
            do_data_transfers (boolean): if we should do data transfers from host to device and back for evaluating
                this function. For better control set this to False and use the method
                ``enqueue_device_access()`` and ``enqueue_host_access`` of the KernelData to set the data.
            is_blocking (boolean): if this is a blocking call, i.e. if we should call finish on all the queues
                after enqueueing the function
            return_events (boolean): if set we also return the last queued events
            wait_for (Dict[CLEnvironment: cl.Event]): per CL environment an event to wait on

        Returns:
            ndarray: the return values of the function, which can be None if this function has a void return type.
                If return_events is set, we return a tuple instead with the results and the last event.
        """
        raise NotImplementedError()

    def get_dependencies(self):
        """Get the list of dependencies this function depends on.

        Returns:
            list[CLFunction]: the list of dependencies for this function.
        """
        raise NotImplementedError()


class SimpleCLCodeObject(CLCodeObject):

    def __init__(self, cl_code):
        """Simple code object for including type definitions in the kernel.

        Args:
            cl_code (str): CL code to be included in the kernel
        """
        self._cl_code = cl_code

    def get_cl_code(self):
        return self._cl_code


class SimpleCLFunction(CLFunction):

    def __init__(self, return_type, cl_function_name, parameter_list, cl_body, dependencies=None, is_kernel_func=False):
        """A simple implementation of a CL function.

        Args:
            return_type (str): the CL return type of the function
            cl_function_name (string): The name of the CL function
            parameter_list (list or tuple): This either contains instances of
                :class:`CLFunctionParameter` or strings from which to form the function parameters.
            cl_body (str): the body of the CL code for this function.
            dependencies (Iterable[CLCodeObject]): The CL code objects this function depends on,
                these will be prepended to the CL code generated by this function.
            is_kernel_func (boolean): if this function should be a kernel function
        """
        super().__init__()
        self._return_type = return_type
        self._function_name = cl_function_name
        self._parameter_list = self._resolve_parameters(parameter_list)
        self._cl_body = cl_body
        self._dependencies = dependencies or []
        self._is_kernel_func = is_kernel_func
        self._compilation_cache = {}

    @classmethod
    def from_string(cls, cl_function, dependencies=()):
        """Parse the given CL function into a SimpleCLFunction object.

        Args:
            cl_function (str): the function we wish to turn into an object
            dependencies (list or tuple of CLLibrary): The list of CL libraries this function depends on

        Returns:
            SimpleCLFunction: the CL data type for this parameter declaration
        """
        is_kernel_func, return_type, function_name, parameter_list, body = split_cl_function(cl_function)
        return SimpleCLFunction(return_type, function_name, parameter_list, body,
                                dependencies=dependencies, is_kernel_func=is_kernel_func)

    def get_cl_function_name(self):
        return self._function_name

    def is_kernel_func(self):
        return self._is_kernel_func

    def get_return_type(self):
        return self._return_type

    def get_parameters(self):
        return self._parameter_list

    def get_kernel_wrapped(self, input_data, nmr_instances, kernel_name=None):
        if self.is_kernel_func():
            return self

        input_data = convert_inputs_to_kernel_data(input_data, self.get_parameters(), nmr_instances)

        kernel_name = kernel_name or 'kernel_' + self.get_cl_function_name()

        variable_inits = []
        function_call_inputs = []
        post_function_callbacks = []
        for parameter in self.get_parameters():
            data = input_data[parameter.name]
            call_args = (parameter.name, '_' + parameter.name, 'gid', parameter.address_space)

            variable_inits.append(data.initialize_variable(*call_args))
            function_call_inputs.append(data.get_function_call_input(*call_args))
            post_function_callbacks.append(data.post_function_callback(*call_args))

        parameter_list = []
        for name, data in input_data.items():
            parameter_list.extend(data.get_kernel_parameters('_' + name))

        assignment = ''
        extra_data = {}
        if self.get_return_type() != 'void':
            assignment = '__return_values[gid] = '
            extra_data = {'__return_values': Zeros((nmr_instances,), self.get_return_type())}
            parameter_list.extend(extra_data['__return_values'].get_kernel_parameters('__return_values'))

        cl_body = '''
            ulong gid = (ulong)(get_global_id(0) / get_local_size(0));

            ''' + '\n'.join(variable_inits) + '''

            ''' + assignment + ' ' + self.get_cl_function_name() + '(' + ', '.join(function_call_inputs) + ''');

            ''' + '\n'.join(post_function_callbacks) + '''
        '''
        func = SimpleCLFunction('void', kernel_name, parameter_list, cl_body,
                                dependencies=[self], is_kernel_func=True)
        return func, extra_data

    def get_signature(self):
        return dedent('{kernel} {return_type} {cl_function_name}({parameters});'.format(
            kernel='kernel' if self.is_kernel_func() else '',
            return_type=self.get_return_type(),
            cl_function_name=self.get_cl_function_name(),
            parameters=', '.join(self._get_parameter_signatures())))

    def get_cl_code(self):
        cl_code = dedent('''
            {kernel} {return_type} {cl_function_name}({parameters}){{
            {body}
            }}
        '''.format(kernel='kernel' if self.is_kernel_func() else '',
                   return_type=self.get_return_type(),
                   cl_function_name=self.get_cl_function_name(),
                   parameters=', '.join(self._get_parameter_signatures()),
                   body=indent(dedent(self._cl_body), ' '*4*4)))

        dependencies = self._get_cl_dependency_code()
        return dependencies + dedent('''
            #ifndef {inclusion_guard_name}
            #define {inclusion_guard_name}
            {code}
            #endif // {inclusion_guard_name}
        '''.format(inclusion_guard_name='INCLUDE_GUARD_{}'.format(self.get_cl_function_name()),
                   code=indent('\n' + cl_code + '\n', ' ' * 4 * 3)))

    def get_cl_body(self):
        return self._cl_body

    def evaluate(self, inputs, nmr_instances, use_local_reduction=False, local_size=None, cl_runtime_info=None,
                 do_data_transfers=True, is_blocking=True, return_events=False, wait_for=None):

        cl_runtime_info = cl_runtime_info or CLRuntimeInfo()

        def resolve_cl_function_and_kernel_data():
            kernel_data = convert_inputs_to_kernel_data(inputs, self.get_parameters(), nmr_instances)
            for data in kernel_data.values():
                data.set_mot_float_dtype(cl_runtime_info.mot_float_dtype)

            cl_function = self
            if not self.is_kernel_func():
                cl_function, extra_data = self.get_kernel_wrapped(kernel_data, nmr_instances)
                kernel_data.update(extra_data)

            return cl_function, kernel_data

        def get_kernel_source(cl_function, kernel_data):
            kernel_source = ''
            kernel_source += get_cl_utility_definitions(cl_runtime_info.double_precision)
            kernel_source += '\n'.join(data.get_type_definitions() for data in kernel_data.values())
            kernel_source += cl_function.get_cl_code()
            return kernel_source

        def get_kernels(kernel_source, function_name):
            hashed_source = hash(kernel_source)
            kernels = {}
            for env in cl_runtime_info.cl_environments:
                key = (hashed_source, env.context, cl_runtime_info.compile_flags)
                if key not in self._compilation_cache:
                    self._compilation_cache[key] = cl.Program(
                        env.context, kernel_source).build(' '.join(cl_runtime_info.compile_flags))
                kernels[env] = getattr(self._compilation_cache[key], function_name)
            return kernels

        cl_function, kernel_data = resolve_cl_function_and_kernel_data()
        kernel_source = get_kernel_source(cl_function, kernel_data)
        kernels = get_kernels(kernel_source, cl_function.get_cl_function_name())

        processor = MultiDeviceProcessor(kernels, kernel_data, cl_runtime_info.cl_environments,
                                         cl_runtime_info.load_balancer, nmr_instances,
                                         use_local_reduction=use_local_reduction,
                                         local_size=local_size, do_data_transfers=do_data_transfers)
        events = processor.process(wait_for=wait_for)

        return_data = None
        if is_blocking:
            processor.finish()
            if self.get_return_type() != 'void':
                return_data = kernel_data['__return_values'].get_data()

        if return_events:
            return return_data, events
        return return_data

    def get_dependencies(self):
        return self._dependencies

    def _get_parameter_signatures(self):
        """Get the signature of the parameters for the CL function declaration.

        This should return the list of signatures of the parameters for use inside the function signature.

        Returns:
            list: the signatures of the parameters for the use in the CL code.
        """
        declarations = []
        for p in self.get_parameters():
            new_p = p.get_renamed(p.name.replace('.', '_'))
            declarations.append(new_p.get_declaration())
        return declarations

    def _get_cl_dependency_code(self):
        """Get the CL code for all the CL code for all the dependencies.

        Returns:
            str: The CL code with the actual code.
        """
        code = ''
        for d in self._dependencies:
            code += d.get_cl_code() + "\n"
        return code

    @staticmethod
    def _resolve_parameters(parameter_list):
        params = []
        for param in parameter_list:
            if isinstance(param, CLFunctionParameter):
                params.append(param)
            else:
                params.append(SimpleCLFunctionParameter(param))
        return params

    def __str__(self):
        return dedent('''
            {return_type} {cl_function_name}({parameters}){{
            {body}
            }}
        '''.format(return_type=self.get_return_type(),
                   cl_function_name=self.get_cl_function_name(),
                   parameters=', '.join(self._get_parameter_signatures()),
                   body=indent(dedent(self._cl_body), ' '*4*4)))

    def __hash__(self):
        return hash(self.__repr__())

    def __eq__(self, other):
        return type(self) == type(other)

    def __ne__(self, other):
        return type(self) != type(other)

    def __call__(self, *args, **kwargs):
        return self.evaluate(*args, **kwargs)


class CLFunctionParameter:

    @property
    def name(self):
        """The name of this parameter.

        Returns:
            str: the name of this parameter
        """
        raise NotImplementedError()

    def get_declaration(self):
        """Get the complete CL declaration for this parameter.

        Returns:
            str: the declaration for this data type.
        """
        raise NotImplementedError()

    @property
    def ctype(self):
        """Get the ctype of this data type.

        For example, if the data type is float4*, we will return float4 here.

        Returns:
            str: the full ctype of this data type
        """
        raise NotImplementedError()

    @property
    def address_space(self):
        """Get the address space of this data declaration.

        Returns:
            str: the data type address space, one of ``global``, ``local``, ``constant`` or ``private``.
        """
        raise NotImplementedError()

    @property
    def basic_ctype(self):
        """Get the basic data type without the vector and pointer additions.

        For example, if the full data ctype is ``float4*``, we will only return ``float`` here.

        Returns:
            str: the raw CL data type
        """
        raise NotImplementedError()

    @property
    def is_vector_type(self):
        """Check if this data type is a vector type (like for example double4, float2, int8, etc.).

        Returns:
            boolean: True if it is a vector type, false otherwise
        """
        raise NotImplementedError()

    @property
    def vector_length(self):
        """Get the length of this vector, returns None if not a vector type.

        Returns:
            int: the length of the vector type (for example, if the data type is float4, this returns 4).
        """
        raise NotImplementedError()

    @property
    def is_pointer_type(self):
        """Check if this parameter is a pointer type (appended by a ``*``)

        Returns:
            boolean: True if it is a pointer type, false otherwise
        """
        raise NotImplementedError()

    @property
    def nmr_pointers(self):
        """Get the number of asterisks / pointer references of this data type.

        If the data type is float**, we return 2 here.

        Returns:
            int: the number of pointer asterisks in the data type.
        """
        raise NotImplementedError()

    @property
    def array_sizes(self):
        """Get the dimension of this array type.

        This returns for example (10, 5) for the data type float[10][5].

        Returns:
            Tuple[int]: the sizes of the arrays
        """
        raise NotImplementedError()

    @property
    def is_array_type(self):
        """Check if this parameter is an array type (like float[3] or int[10][5]).

        Returns:
            boolean: True if this is an array type, false otherwise
        """
        raise NotImplementedError()

    def get_renamed(self, name):
        """Get a copy of the current parameter but then with a new name.

        Args:
            name (str): the new name for this parameter

        Returns:
            cls: a copy of the current type but with a new name
        """
        raise NotImplementedError()


_cl_data_type_parser = tatsu.compile('''
    result = [address_space] {type_qualifiers}* ctype {pointer_star}* {pointer_qualifiers}* name {array_size}*;

    address_space = ['__'] ('local' | 'global' | 'constant' | 'private');
    type_qualifiers = 'const' | 'volatile';

    basic_ctype = ?'(unsigned )?\w[\w]*[a-zA-Z]';
    vector_type_length = '2' | '3' | '4' | '8' | '16';
    ctype = basic_ctype [vector_type_length];
    pointer_star = '*';

    pointer_qualifiers = 'const' | 'restrict';

    name = /[\w\_\-\.]+/;
    array_size = /\[\d+\]/;
''')


class SimpleCLFunctionParameter(CLFunctionParameter):

    def __init__(self, declaration):
        """Creates a new function parameter for the CL functions.

        Args:
            declaration (str): the declaration of this parameter. For example ``global int foo``.
        """
        self._address_space = None
        self._type_qualifiers = []
        self._basic_ctype = ''
        self._vector_type_length = None
        self._nmr_pointer_stars = 0
        self._pointer_qualifiers = []
        self._name = ''
        self._array_sizes = []

        param = self

        class Semantics:

            def type_qualifiers(self, ast):
                if ast in param._type_qualifiers:
                    raise ValueError('The pre-type qualifier "{}" is present multiple times.'.format(ast))
                param._type_qualifiers.append(ast)
                return ast

            def address_space(self, ast):
                param._address_space = ''.join(ast)
                return ''.join(ast)

            def basic_ctype(self, ast):
                param._basic_ctype = ast
                return ast

            def vector_type_length(self, ast):
                param._vector_type_length = int(ast)
                return ast

            def pointer_star(self, ast):
                param._nmr_pointer_stars += 1
                return ast

            def pointer_qualifiers(self, ast):
                if ast in param._pointer_qualifiers:
                    raise ValueError('The pre-type qualifier "{}" is present multiple times.'.format(ast))
                param._pointer_qualifiers.append(ast)
                return ast

            def name(self, ast):
                param._name = ast
                return ast

            def array_size(self, ast):
                param._array_sizes.append(int(ast[1:-1]))
                return ast

        _cl_data_type_parser.parse(declaration, semantics=Semantics())

    @property
    def name(self):
        return self._name

    def get_renamed(self, name):
        new_param = copy(self)
        new_param._name = name
        return new_param

    def get_declaration(self):
        declaration = ''

        if self._address_space:
            declaration += str(self._address_space) + ' '

        if self._type_qualifiers:
            declaration += str(' '.join(self._type_qualifiers)) + ' '

        declaration += str(self.ctype)
        declaration += '*' * self._nmr_pointer_stars

        if self._pointer_qualifiers:
            declaration += ' ' + str(' '.join(self._pointer_qualifiers)) + ' '

        declaration += ' ' + self._name

        for s in self._array_sizes:
            declaration += '[{}]'.format(s)

        return declaration

    @property
    def ctype(self):
        if self._vector_type_length is not None:
            return '{}{}'.format(self._basic_ctype, self._vector_type_length)
        return self._basic_ctype

    @property
    def address_space(self):
        return self._address_space or 'private'

    @property
    def basic_ctype(self):
        return self._basic_ctype

    @property
    def is_vector_type(self):
        return self._vector_type_length is not None

    @property
    def vector_length(self):
        return self._vector_type_length

    @property
    def is_pointer_type(self):
        return self._nmr_pointer_stars > 0

    @property
    def nmr_pointers(self):
        return self._nmr_pointer_stars

    @property
    def array_sizes(self):
        return self._array_sizes

    @property
    def is_array_type(self):
        return len(self.array_sizes) > 0



import numbers
from collections import OrderedDict, Mapping

import numpy as np
import pyopencl as cl

from mot.lib.cl_environments import CLEnvironment
from mot.lib.utils import dtype_to_ctype, ctype_to_dtype, convert_data_to_dtype, is_vector_ctype, split_vector_ctype

__author__ = 'Robbert Harms'
__date__ = '2018-04-09'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert@xkls.nl'
__licence__ = 'LGPL v3'


class KernelData:

    @property
    def ctype(self):
        """Get the ctype of this kernel data element

        Returns:
            str: the ctype of this kernel data element
        """
        raise NotImplementedError()

    def get_subset(self, problem_indices=None, batch_range=None):
        """Get a subset of this kernel data over problem instances.

        This can be used to get a subset of a kernel data object such that only a subset of the problem instances
        are processed.

        One can either provide a list of indices or a range defined by a start and end position. Providing a list
        of indices is more flexible, but will incur a memory copy. Providing a range only allows contiguous slices,
        but it will use the original memory.

        Args:
            problem_indices (Iterable or None): list of problem instances we would like to have in the subset.
                May return ``self``. If problem_indices evaluates to None, this should return ``self``.
            batch_range (Tuple(int, int)): the range specified by a starting and ending position. The
                Start position is taken inclusive, the end position is taken exclusive.

        Returns:
            KernelData: a kernel data object of the same type as the current kernel data object.
        """
        raise NotImplementedError()

    def set_mot_float_dtype(self, mot_float_dtype):
        """Set the numpy data type corresponding to the ``mot_float_type`` ctype.

        This is set just prior to using this kernel data in the kernel.

        Args:
            mot_float_dtype (dtype): the numpy data type that is to correspond with the ``mot_float_type`` used in the
                kernels.
        """
        raise NotImplementedError()

    def get_data(self):
        """Get the underlying data of this kernel data object.

        Returns:
            dict, ndarray, scalar: the underlying data object, can return None if this input data has no actual data.
        """
        raise NotImplementedError()

    def get_children(self):
        """Get a list of children kernel data elements.

        Returns:
            List[KernelData]: list of children kernel data elements, can be an an empty list.
        """
        raise NotImplementedError()

    def get_scalar_arg_dtypes(self):
        """Get the numpy data types we should report in the kernel call for scalar elements.

        If we are inserting scalars in the kernel we need to provide the CL runtime with the correct data type
        of the function. If the kernel parameter is not a scalar, we should return None. If the kernel data does not
        require a kernel input parameter, return an empty list.

        This list should match the list of parameters of :meth:`get_kernel_parameters`.

        Returns:
            List[Union[dtype, None]]: the numpy data type for this element, or None if this is not a scalar.
        """
        raise NotImplementedError()

    def get_type_definitions(self):
        """Get possible type definitions needed to load this data into the kernel.

        These types are defined at the head of the CL script, before any functions.

        Returns:
            str: a CL compatible type declaration. This can for example be used for defining struct types.
                If no extra types are needed, this function should return the empty string.
        """
        raise NotImplementedError()

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        """Initialize the variable inside the kernel function.

        This should initialize the variable as such that we can use it when calling the function acting on this data.

        Args:
             variable_name (str): the name for this variable
             kernel_param_name (str): the kernel parameter name (given in :meth:`get_kernel_parameters`).
             problem_id_substitute (str): the substitute for the ``{problem_id}`` in the kernel data info elements.
             address_space (str): the desired address space for this variable, defined by the parameter of the called
                function.

        Returns:
            str: the necessary CL code to initialize this variable
        """
        raise NotImplementedError()

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        """How this kernel data is used as input to the function that operates on the data.

        Args:
            variable_name (str): the name for this variable
            kernel_param_name (str): the kernel parameter name (given in :meth:`get_kernel_parameters`).
            problem_id_substitute (str): the substitute for the ``{problem_id}`` in the kernel data info elements.
            address_space (str): the desired address space for this variable, defined by the parameter of the called
                function.

        Returns:
            str: a single string representing how this kernel data is used as input to the function we are applying
        """
        raise NotImplementedError()

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        """A callback to update or change data after the function has been applied

        Args:
            variable_name (str): the name for this variable
            kernel_param_name (str): the kernel parameter name (given in :meth:`get_kernel_parameters`).
            problem_id_substitute (str): the substitute for the ``{problem_id}`` in the kernel data info elements.
            address_space (str): the desired address space for this variable, defined by the parameter of the called
                function.

        Returns:
            str: CL code that needs to be run after the function has been applied.
        """
        raise NotImplementedError()

    def get_struct_declaration(self, name):
        """Get the variable declaration of this data object for use in a Struct.

        Args:
            name (str): the name for this data

        Returns:
            str: the variable declaration of this kernel data object
        """
        raise NotImplementedError()

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        """Initialize the variable inside a struct.

        This should initialize the variable for use in a struct (should correspond to :meth:`get_struct_declaration`).

        Args:
             variable_name (str): the name for this variable
             kernel_param_name (str): the kernel parameter name (given in :meth:`get_kernel_parameters`).
             problem_id_substitute (str): the substitute for the ``{problem_id}`` in the kernel data info elements.

        Returns:
            str: the necessary CL code to initialize this variable
        """
        raise NotImplementedError()

    def get_kernel_parameters(self, kernel_param_name):
        """Get the kernel argument declarations for this kernel data.

        Args:
            kernel_param_name (str): the parameter name for the parameter in the kernel function call

        Returns:
            List[str]: a list of kernel parameter declarations, or an empty list
        """
        raise NotImplementedError()

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        """Enqueue either a map or write operation for this kernel input data object.

        This should add non-blocking maps or write operations to the given queue.

        Args:
            cl_environments (mot.lib.cl_environments.CLEnvironment or list):
                enqueue host access for one or multiple CL environments
            is_blocking (boolean): if the enqueuing should be blocking calls or not
                This first enqueues all the work to all the contexts and then waits for completion on them.
            wait_for (Dict[CLEnvironment: cl.Event]): per CL environment an event to wait on

        Returns:
            Dict[CLEnvironment: cl.Event]: per CLEnvironment an event for waiting on
        """
        raise NotImplementedError()

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        """Enqueue either an unmap or read operation for this kernel input data object.

        This should add non-blocking unmaps or read operations to the given queue.

        Args:
            cl_environments (mot.lib.cl_environments.CLEnvironment or list):
                enqueue device access for one or multiple CL environments
            is_blocking (boolean): if the enqueuing should be blocking calls or not
                This first enqueues all the work to all the contexts and then waits for completion on them.
            wait_for (Dict[CLEnvironment: cl.Event]): per CL environment an event to wait on

        Returns:
            Dict[CLEnvironment: cl.Event]: per CLEnvironment an event for waiting on
        """
        raise NotImplementedError()

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        """Get the kernel input data matching the list of parameters of :meth:`get_kernel_parameters`.

        Since the kernels follow the map/unmap paradigm make sure to use the ``USE_HOST_PTR`` when making
        writable data objects.

        Args:
            cl_environment (mot.lib.cl_environments.CLEnvironment): the environment for which to prepare the input
            workgroup_size (int): the workgroup size the kernel will use.

        Returns:
            List: a list of buffers, local memory objects, scalars, etc., anything that can be loaded into the kernel.
                If no data should be entered, return an empty list.
        """
        raise NotImplementedError()

    def get_nmr_kernel_inputs(self):
        """Get the number of kernel inputs this input data object has.

        Returns:
            int: the number of kernel inputs
        """
        raise NotImplementedError()


class Struct(KernelData):

    def __init__(self, elements, ctype, anonymous=False):
        """A kernel data element for structs.

        Please be aware that structs will always be passed as a pointer to the calling function.

        Args:
            elements (Dict[str, Union[Dict, KernelData]]): the kernel data elements to load into the kernel
                Can be a nested dictionary, in which case we load the nested elements as anonymous structs.
                Alternatively, you can nest Structs in Structs, yielding named structs.
            ctype (str): the name of this structure
            anonymous (boolean): if this struct is to be loaded anonymously, this is only meant for nested Structs.
        """
        self._elements = OrderedDict(sorted(elements.items()))

        for key in list(self._elements):
            value = self._elements[key]
            if isinstance(value, Mapping):
                self._elements[key] = Struct(value, key, anonymous=True)

        self._ctype = ctype
        self._anonymous = anonymous

    @property
    def ctype(self):
        return self._ctype

    def get_subset(self, problem_indices=None, batch_range=None):
        if problem_indices is None and batch_range is None:
            return self
        sub_elements = OrderedDict([(k, v.get_subset(problem_indices, batch_range))
                                    for k, v in self._elements.items()])
        return Struct(sub_elements, self._ctype, anonymous=self._anonymous)

    def set_mot_float_dtype(self, mot_float_dtype):
        for element in self._elements.values():
            element.set_mot_float_dtype(mot_float_dtype)

    def get_data(self):
        data = {}
        for name, value in self._elements.items():
            data[name] = value.get_data()
        return data

    def get_children(self):
        return self._elements.values()

    def get_scalar_arg_dtypes(self):
        dtypes = []
        for d in self._elements.values():
            dtypes.extend(d.get_scalar_arg_dtypes())
        return dtypes

    def get_type_definitions(self):
        other_structs = '\n'.join(element.get_type_definitions() for element in self._elements.values())

        if self._anonymous:
            return other_structs

        return other_structs + '''
            #ifndef {inclusion_guard_name}
            #define {inclusion_guard_name}
            typedef struct {ctype}{{
                {definitions}
            }} {ctype};
            #endif // {inclusion_guard_name}
        '''.format(ctype=self._ctype,
                   definitions='\n'.join(data.get_struct_declaration(name) for name, data in self._elements.items()),
                   inclusion_guard_name='INCLUDE_GUARD_CL_EXTRA_{}'.format(self._ctype))

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return_str = ''
        for name, data in self._elements.items():
            return_str += data.initialize_variable('{}_{}'.format(variable_name, name),
                                                   '{}_{}'.format(kernel_param_name, name),
                                                   problem_id_substitute, 'global')

        inits = [data.get_struct_initialization(
            '{}_{}'.format(variable_name, name),
            '{}_{}'.format(kernel_param_name, name), problem_id_substitute) for name, data in self._elements.items()]

        if self._anonymous:
            return return_str + '''
                struct {ctype}{{
                    {definitions}
                }};
                struct {ctype} {v_name} = {{ {inits} }};
            '''.format(ctype=self._ctype, v_name=variable_name, inits=', '.join(inits),
                       definitions='\n'.join(data.get_struct_declaration(name)
                                             for name, data in self._elements.items()))

        return return_str + '''
            {ctype} {v_name} = {{ {inits} }};
        '''.format(ctype=self._ctype, v_name=variable_name, inits=', '.join(inits))

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return '&' + variable_name

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return ''

    def get_struct_declaration(self, name):
        if self._anonymous:
            return '''
                struct {{
                    {definitions}
                }}* {name};
            '''.format(
                name=name,
                definitions='\n'.join(data.get_struct_declaration(name) for name, data in self._elements.items()))
        return '{}* {};'.format(self._ctype, name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        if self._anonymous:
            return '(void*)(&{})'.format(variable_name)
        return '&' + variable_name

    def get_kernel_parameters(self, kernel_param_name):
        parameters = []
        for name, d in self._elements.items():
            parameters.extend(d.get_kernel_parameters('{}_{}'.format(kernel_param_name, name)))
        return parameters

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        events = {}
        for d in self._elements.values():
            events.update(d.enqueue_host_access(cl_environments, is_blocking=is_blocking, wait_for=wait_for))
        return events

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        events = {}
        for d in self._elements.values():
            events.update(d.enqueue_device_access(cl_environments, is_blocking=is_blocking, wait_for=wait_for))
        return events

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        data = []
        for d in self._elements.values():
            data.extend(d.get_kernel_inputs(cl_environment, workgroup_size))
        return data

    def get_nmr_kernel_inputs(self):
        return sum(element.get_nmr_kernel_inputs() for element in self._elements.values())

    def __getitem__(self, key):
        return self._elements[key]

    def __contains__(self, key):
        return key in self._elements

    def __len__(self):
        return len(self._elements)


class Scalar(KernelData):

    def __init__(self, value, ctype=None, inline=True):
        """A kernel input scalar.

        This will insert the given value directly into the kernel's source code, and will not load it as a buffer.

        Args:
            value (number): the number to insert into the kernel as a scalar.
            ctype (str): the desired c-type for in use in the kernel, like ``int``, ``float`` or ``mot_float_type``.
                If None it is implied from the value.
            inline (bool): if set to True, we inline the value in the generated kernel code,
                else we set it as an argument.
        """
        if isinstance(value, str) and value == 'INFINITY':
            self._value = np.inf
        elif isinstance(value, str) and value == '-INFINITY':
            self._value = -np.inf
        else:
            self._value = np.array(value)
        self._ctype = ctype or dtype_to_ctype(self._value.dtype)
        self._mot_float_dtype = None
        self._inline = inline

    @property
    def ctype(self):
        return self._ctype

    def get_subset(self, problem_indices=None, batch_range=None):
        return self

    def get_data(self):
        if self._ctype.startswith('mot_float_type'):
            return np.asscalar(self._value.astype(self._mot_float_dtype))
        return np.asscalar(self._value)

    def get_children(self):
        return []

    def get_scalar_arg_dtypes(self):
        if self._inline:
            return []
        return [ctype_to_dtype(self._ctype)]

    def get_type_definitions(self):
        return ''

    def get_struct_declaration(self, name):
        return '{} {};'.format(self._ctype, name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        if self._inline:
            return self.get_function_call_input(variable_name, kernel_param_name, problem_id_substitute, 'private')
        return kernel_param_name

    def get_kernel_parameters(self, kernel_param_name):
        if self._inline:
            return []
        return ['{} {}'.format(self._ctype, kernel_param_name)]

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        if self._inline:
            return []
        return [self.get_data()]

    def get_nmr_kernel_inputs(self):
        if self._inline:
            return 0
        return 1

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return ''

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if not self._inline:
            return kernel_param_name

        if is_vector_ctype(self._ctype):
            vector_length = split_vector_ctype(self._ctype)[1]

            values = [str(el) for el in np.atleast_1d(np.squeeze(self._value))]

            if len(values) < vector_length:
                values.extend(['0'] * (vector_length - len(values)))
            assignment = '(' + self._ctype + ')(' + ', '.join(values) + ')'

        elif np.isposinf(self._value):
            assignment = 'INFINITY'
        elif np.isneginf(self._value):
            assignment = '-INFINITY'
        else:
            assignment = str(np.squeeze(self._value))
        return assignment

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return ''


class PrivateMemory(KernelData):

    def __init__(self, nmr_items, ctype):
        """Adds a private memory array of the indicated size to the kernel data elements.

        This is useful if you want to have private memory arrays in kernel data structs.

        Args:
            nmr_items (int): the size of the private memory array
            ctype (str): the desired c-type for this local memory object, like ``int``, ``float`` or ``mot_float_type``.
        """
        self._ctype = ctype
        self._mot_float_dtype = None
        self._nmr_items = nmr_items

    @property
    def ctype(self):
        return self._ctype

    def get_subset(self, problem_indices=None, batch_range=None):
        return self

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

    def get_data(self):
        return None

    def get_children(self):
        return []

    def get_scalar_arg_dtypes(self):
        return []

    def get_type_definitions(self):
        return ''

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return '''
            {ctype} {v_name}[{nmr_elements}];
        '''.format(ctype=self._ctype, v_name=kernel_param_name, nmr_elements=self._nmr_items)

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return kernel_param_name

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return ''

    def get_struct_declaration(self, name):
        return '{}* {};'.format(self._ctype, name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        return kernel_param_name

    def get_kernel_parameters(self, kernel_param_name):
        return []

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        return []

    def get_nmr_kernel_inputs(self):
        return 0


class LocalMemory(KernelData):

    def __init__(self, ctype, nmr_items=None):
        """Indicates that a local memory array of the indicated size must be loaded as kernel input data.

        By default, this will create a local memory object the size of the local work group.

        Args:
            ctype (str): the desired c-type for this local memory object, like ``int``, ``float`` or ``mot_float_type``.
            nmr_items (int or Callable[[int], int]): either the size directly or a function that can calculate the
                required local memory size given the work group size. This will independently be multiplied with the
                item size of the ctype for the final size in bytes.
        """
        self._ctype = ctype
        self._mot_float_dtype = None

        if nmr_items is None:
            self._size_func = lambda workgroup_size: workgroup_size
        elif isinstance(nmr_items, numbers.Number):
            self._size_func = lambda _: nmr_items
        else:
            self._size_func = nmr_items

    @property
    def ctype(self):
        return self._ctype

    def get_subset(self, problem_indices=None, batch_range=None):
        return self

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

    def get_data(self):
        return None

    def get_children(self):
        return []

    def get_scalar_arg_dtypes(self):
        return [None]

    def get_type_definitions(self):
        return ''

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return ''

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return kernel_param_name

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return ''

    def get_struct_declaration(self, name):
        return 'local {}* restrict {};'.format(self._ctype, name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        return self.get_function_call_input(variable_name, kernel_param_name, problem_id_substitute, '')

    def get_kernel_parameters(self, kernel_param_name):
        return ['local {}* restrict {}'.format(self._ctype, kernel_param_name)]

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        mot_float_type_dtype = None
        if self._mot_float_dtype:
            mot_float_type_dtype = dtype_to_ctype(self._mot_float_dtype)

        itemsize = np.dtype(ctype_to_dtype(self._ctype, mot_float_type_dtype)).itemsize
        return [cl.LocalMemory(itemsize * self._size_func(workgroup_size))]

    def get_nmr_kernel_inputs(self):
        return 1


class Array(KernelData):

    def __init__(self, data, ctype=None, as_scalar=False, parallelize_over_first_dimension=True,
                 mode='rw', use_host_ptr=True):
        """Loads the given array as a buffer into one or more OpenCL contexts.

        By default, this expects multi-dimensional arrays (n, m, k, ...) which holds a (m, k, ...) for every data
        point ``n``. As such, every global work item will only receive a subset of the matrix. If every work item
        needs access to all items, set parallelize_over_first_dimension to False.

        In general, this class uses the provided data's dtype for the buffering. If the ctype is provided and the
        data's dtype does not match the ctype, we will convert the data to the righ dtype. If the ctype is set
        to mot_float_type and the mot_float_type changes, then a new Array data class is returned.

        Args:
            data (ndarray): the data to load in the kernel
            ctype (str): the desired c-type for in use in the kernel, like ``int``, ``float`` or ``mot_float_type``.
                If None it is implied from the provided data. If not matching the data, we convert the data.
            as_scalar (boolean): if given and if the data is only a 1d, we will load the value as a scalar in the
                data struct. As such, one does not need to evaluate as a pointer.
            parallelize_over_first_dimension (boolean): only applicable for multi-dimensional arrays (n, m, k, ...)
                where ``n`` is expected to correspond to problem instances. If True, we will load the data as
                (m, k, ...) arrays for each problem instance. If False, the data will be loaded as is, and each problem
                instance will have a reference to the complete array.
            mode (str): one of 'r', 'w' or 'rw', for respectively read, write or read and write.
                This defines how the device is planned on accessing this array, and, defines if we will read the
                data back after applying a kernel (if 'w' is included we will write data back, else, not).
            use_host_ptr (boolean): if set, we will use the USE_HOST_PTR flag and use map/unmap for data transfers
                if not set, we create a device side buffer and use explicit read and write commands to transfer the data
        """
        if isinstance(data, (list, tuple)):
            data = np.array(data)

        self._mode = mode
        self._is_readable = 'r' in mode
        self._is_writable = 'w' in mode

        self._data = data
        if ctype and not ctype.startswith('mot_float_type'):
            self._data = convert_data_to_dtype(self._data, ctype)

        self._ctype = ctype or dtype_to_ctype(self._data.dtype)
        self._mot_float_dtype = None
        self._backup_data_reference = None
        self._as_scalar = as_scalar
        self._parallelize_over_first_dimension = parallelize_over_first_dimension

        self._use_host_ptr = use_host_ptr
        self._buffer_cache = {}  # caching the buffers per context

        self._data_length = 1
        if len(self._data.shape):
            self._data_length = self._data.strides[0] // self._data.itemsize
        if not self._parallelize_over_first_dimension:
            self._data_length = self._data.size

        if self._as_scalar and len(np.squeeze(self._data).shape) > 1:
            raise ValueError('The option "as_scalar" was set, but the data has more than one dimensions.')

    @property
    def ctype(self):
        return self._ctype

    @property
    def mode(self):
        """Get the read write mode defined for this array.

        Returns:
            str: the mode in which the buffer should be created, a string like 'r', 'w' or 'rw'
        """
        return self._mode

    def get_subset(self, problem_indices=None, batch_range=None):
        if problem_indices is None and batch_range is None:
            return self
        if not self._parallelize_over_first_dimension:
            return self

        if problem_indices is None:
            return Array(self._data[batch_range[0]:batch_range[1]], ctype=self._ctype,
                         mode=self._mode, as_scalar=self._as_scalar,
                         parallelize_over_first_dimension=self._parallelize_over_first_dimension,
                         use_host_ptr=self._use_host_ptr)

        def is_consecutive(l):
            return np.sum(np.diff(np.sort(l)) == 1) >= (len(l) - 1)

        if is_consecutive(problem_indices):
            return Array(self._data[problem_indices[0]:(problem_indices[-1] + 1)], ctype=self._ctype,
                         mode=self._mode, as_scalar=self._as_scalar,
                         parallelize_over_first_dimension=self._parallelize_over_first_dimension,
                         use_host_ptr=self._use_host_ptr)

        return Array(self._data[problem_indices], ctype=self._ctype,
                     mode=self._mode, as_scalar=self._as_scalar,
                     parallelize_over_first_dimension=self._parallelize_over_first_dimension,
                     use_host_ptr=self._use_host_ptr)

    def set_mot_float_dtype(self, mot_float_dtype):
        self._mot_float_dtype = mot_float_dtype

        if self._ctype.startswith('mot_float_type'):
            if self._backup_data_reference is not None:
                self._data = self._backup_data_reference
                self._backup_data_reference = None

            new_data = convert_data_to_dtype(self._data, self._ctype,
                                             mot_float_type=dtype_to_ctype(mot_float_dtype))

            if new_data is not self._data:
                self._backup_data_reference = self._data
                self._data = new_data
                self._buffer_cache = {}  # cache is invalidated

        # data length may change when an CL vector type is converted from (n, 3) shape to (n,)
        self._data_length = 1
        if len(self._data.shape):
            self._data_length = self._data.strides[0] // self._data.itemsize
        if not self._parallelize_over_first_dimension:
            self._data_length = self._data.size

    def get_data(self):
        return self._data

    def get_children(self):
        return []

    def get_scalar_arg_dtypes(self):
        return [None]

    def get_type_definitions(self):
        return ''

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if not self._as_scalar:
            if address_space == 'private':
                return '''
                    private {ctype} {v_name}[{nmr_elements}];

                    for(uint i = 0; i < {nmr_elements}; i++){{
                        {v_name}[i] = {k_name}[{offset} + i];
                    }}
                '''.format(ctype=self._ctype, v_name=variable_name, k_name=kernel_param_name,
                           nmr_elements=self._data_length,
                           offset=self._get_offset_str(problem_id_substitute))
            elif address_space == 'local':
                return '''
                    local {ctype} {v_name}[{nmr_elements}];

                    if(get_local_id(0) == 0){{
                        for(uint i = 0; i < {nmr_elements}; i++){{
                            {v_name}[i] = {k_name}[{offset} + i];
                        }}
                    }}
                    barrier(CLK_LOCAL_MEM_FENCE);
                '''.format(ctype=self._ctype, v_name=variable_name, k_name=kernel_param_name,
                           nmr_elements=self._data_length,
                           offset=self._get_offset_str(problem_id_substitute))
        return ''

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if self._as_scalar:
            return '{}[{}]'.format(kernel_param_name, self._get_offset_str(problem_id_substitute))
        else:
            if address_space == 'global':
                return '{} + {}'.format(kernel_param_name, self._get_offset_str(problem_id_substitute))
            elif address_space == 'private':
                return variable_name
            elif address_space == 'local':
                return variable_name

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if self._is_writable:
            if not self._as_scalar:
                if address_space == 'private':
                    return '''
                        for(uint i = 0; i < {nmr_elements}; i++){{
                            {k_name}[{offset} + i] = {v_name}[i];
                        }}
                    '''.format(v_name=variable_name, k_name=kernel_param_name,
                               nmr_elements=self._data_length, offset=self._get_offset_str(problem_id_substitute))
                elif address_space == 'local':
                    return '''
                        if(get_local_id(0) == 0){{
                            for(uint i = 0; i < {nmr_elements}; i++){{
                                {k_name}[{offset} + i] = {v_name}[i];
                            }}
                        }}
                    '''.format(v_name=variable_name, k_name=kernel_param_name,
                               nmr_elements=self._data_length,
                               offset=self._get_offset_str(problem_id_substitute))
        return ''

    def get_struct_declaration(self, name):
        if self._as_scalar:
            return '{} {};'.format(self._ctype, name)
        return 'global {}* restrict {};'.format(self._ctype, name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        return self.get_function_call_input(variable_name, kernel_param_name, problem_id_substitute, 'global')

    def get_kernel_parameters(self, kernel_param_name):
        return ['global {}* restrict {}'.format(self._ctype, kernel_param_name)]

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        if isinstance(cl_environments, CLEnvironment):
            cl_environments = [cl_environments]

        for env in cl_environments:
            self.get_kernel_inputs(env, 1)

        wait_for = wait_for or {}
        events = {}

        if self._is_writable:
            for env in cl_environments:
                context = env.context

                wait_list = []
                for wait_env, wait_event in wait_for.items():
                    if wait_env.context is context:
                        wait_list.append(wait_event)

                if not any(e.context is env.context for e in events.keys()):
                    if self._use_host_ptr:
                        _, event = cl.enqueue_map_buffer(
                            env.queue, self._buffer_cache[context],
                            cl.map_flags.READ, 0, self._data.shape, self._data.dtype,
                            order="C", wait_for=wait_list, is_blocking=False)
                    else:
                        event = cl.enqueue_copy(env.queue, self._data, self._buffer_cache[context],
                                                is_blocking=False, wait_for=wait_list)

                    events[env] = event

        if is_blocking:
            for env in cl_environments:
                env.queue.finish()

        return events

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        if isinstance(cl_environments, CLEnvironment):
            cl_environments = [cl_environments]

        for env in cl_environments:
            self.get_kernel_inputs(env, 1)

        wait_for = wait_for or {}
        events = {}

        if not self._use_host_ptr:
            for env in cl_environments:
                context = env.context

                wait_list = []
                for wait_env, wait_event in wait_for.items():
                    if wait_env.context is context:
                        wait_list.append(wait_event)

                if not any(e.context is env.context for e in events.keys()):
                    event = cl.enqueue_copy(env.queue, self._buffer_cache[context], self._data,
                                            is_blocking=False, wait_for=wait_list)
                    events[env] = event

        if is_blocking:
            for env in cl_environments:
                env.queue.finish()

        return events

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        def get_mem_flags():
            if self._is_writable:
                if self._is_readable:
                    return cl.mem_flags.READ_WRITE
                else:
                    return cl.mem_flags.WRITE_ONLY
            else:
                return cl.mem_flags.READ_ONLY

        cl_context = cl_environment.context

        if cl_context not in self._buffer_cache:
            if self._use_host_ptr:
                self._buffer_cache[cl_context] = cl.Buffer(cl_context,
                                                           get_mem_flags() | cl.mem_flags.USE_HOST_PTR,
                                                           hostbuf=self._data)
            else:
                self._buffer_cache[cl_context] = cl.Buffer(cl_context, get_mem_flags(), size=self._data.nbytes)

        return [self._buffer_cache[cl_context]]

    def get_nmr_kernel_inputs(self):
        return 1

    def _get_offset_str(self, problem_id_substitute):
        if self._parallelize_over_first_dimension:
            offset_str = str(self._data_length) + ' * {problem_id}'
        else:
            offset_str = '0'
        return offset_str.replace('{problem_id}', problem_id_substitute)


class Zeros(KernelData):

    def __init__(self, shape, ctype, parallelize_over_first_dimension=True, host_accessible=True, mode='rw'):
        """Allocate an output buffer of the given shape.

        This is meant to quickly allocate a buffer large enough to hold the data requested. After running an OpenCL
        kernel you can get the written data using the method :meth:`get_data`.

        If host accessible is False and you call the method :meth:`get_subset`, the resulting computations will
        be split over multiple buffers and there is no single continuous buffer with all the values. Therefore,
        this flag is not usable if you intend to share a :class:`Zeros` with multiple CLEnvironments.

        Args:
            shape (int or tuple): the shape of the output array
            ctype (str): the desired C-type for this zero's array
            parallelize_over_first_dimension (boolean): only applicable for multi-dimensional arrays (n, m, k, ...)
                where ``n`` is expected to correspond to problem instances. If True, we will load the data as
                (m, k, ...) arrays for each problem instance. If False, the data will be loaded as is, and each problem
                instance will have a reference to the complete array.
            host_accessible (boolean): if the array needs to be accessible from the host. If not, we can allocate
                the buffer on the device, saving time on memory copies.
            mode (str): one of 'r', 'w' or 'rw', for respectively read, write or read and write. This sets the
                mode of how the data is loaded into the compute device's memory.
        """
        self._shape = shape
        if isinstance(self._shape, numbers.Number):
            self._shape = (self._shape,)
        self._host_accessible = host_accessible
        self._ctype = ctype

        self._mode = mode
        self._is_readable = 'r' in mode
        self._is_writable = 'w' in mode

        self._mot_float_dtype = None
        self._parallelize_over_first_dimension = parallelize_over_first_dimension

        self._buffer_cache = {}  # caching the buffers per context

        self._data_length = 1
        if isinstance(shape, (list, tuple)) and len(shape) > 1:
            self._data_length = np.prod(shape[1:])
        if not self._parallelize_over_first_dimension:
            self._data_length = np.prod(shape)

        if host_accessible:
            self._array = Array(np.zeros(shape, dtype=ctype_to_dtype(ctype)), ctype,
                                parallelize_over_first_dimension=parallelize_over_first_dimension,
                                mode=mode, as_scalar=False)

    @property
    def ctype(self):
        return self._ctype

    @property
    def mode(self):
        return self._mode

    def get_subset(self, problem_indices=None, batch_range=None):
        if self._host_accessible:
            return self._array.get_subset(problem_indices, batch_range)
        if problem_indices is not None:
            shape = (len(problem_indices),) + self._shape[1:]
        else:
            shape = (batch_range[1] - batch_range[0],) + self._shape[1:]
        return Zeros(shape, self._ctype, mode=self._mode,
                     parallelize_over_first_dimension=self._parallelize_over_first_dimension,
                     host_accessible=False)

    def set_mot_float_dtype(self, mot_float_dtype):
        if self._host_accessible:
            return self._array.set_mot_float_dtype(mot_float_dtype)
        self._mot_float_dtype = mot_float_dtype

    def get_data(self):
        if self._host_accessible:
            return self._array.get_data()
        raise ValueError('Can not get the data from a non host accessible array.')

    def get_children(self):
        return []

    def get_scalar_arg_dtypes(self):
        return [None]

    def get_type_definitions(self):
        return ''

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if self._host_accessible:
            return self._array.initialize_variable(variable_name, kernel_param_name,
                                                   problem_id_substitute, address_space)

        if address_space == 'private':
            return '''
                private {ctype} {v_name}[{nmr_elements}];

                for(uint i = 0; i < {nmr_elements}; i++){{
                    {v_name}[i] = {k_name}[{offset} + i];
                }}
            '''.format(ctype=self._ctype, v_name=variable_name, k_name=kernel_param_name,
                       nmr_elements=self._data_length,
                       offset=self._get_offset_str(problem_id_substitute))
        elif address_space == 'local':
            return '''
                local {ctype} {v_name}[{nmr_elements}];

                if(get_local_id(0) == 0){{
                    for(uint i = 0; i < {nmr_elements}; i++){{
                        {v_name}[i] = {k_name}[{offset} + i];
                    }}
                }}
                barrier(CLK_LOCAL_MEM_FENCE);
            '''.format(ctype=self._ctype, v_name=variable_name, k_name=kernel_param_name,
                       nmr_elements=self._data_length,
                       offset=self._get_offset_str(problem_id_substitute))

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if self._host_accessible:
            return self._array.get_function_call_input(variable_name, kernel_param_name,
                                                       problem_id_substitute, address_space)

        if address_space == 'global':
            return '{} + {}'.format(kernel_param_name, self._get_offset_str(problem_id_substitute))
        elif address_space == 'private':
            return variable_name
        elif address_space == 'local':
            return variable_name

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        if self._host_accessible:
            return self._array.post_function_callback(variable_name, kernel_param_name,
                                                      problem_id_substitute, address_space)

        if self._is_writable:
            if address_space == 'private':
                return '''
                    for(uint i = 0; i < {nmr_elements}; i++){{
                        {k_name}[{offset} + i] = {v_name}[i];
                    }}
                '''.format(v_name=variable_name, k_name=kernel_param_name,
                           nmr_elements=self._data_length, offset=self._get_offset_str(problem_id_substitute))
            elif address_space == 'local':
                return '''
                    if(get_local_id(0) == 0){{
                        for(uint i = 0; i < {nmr_elements}; i++){{
                            {k_name}[{offset} + i] = {v_name}[i];
                        }}
                    }}
                '''.format(v_name=variable_name, k_name=kernel_param_name,
                           nmr_elements=self._data_length,
                           offset=self._get_offset_str(problem_id_substitute))
        return ''

    def get_struct_declaration(self, name):
        if self._host_accessible:
            return self._array.get_struct_declaration(name)

        return 'global {}* restrict {};'.format(self._ctype, name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        if self._host_accessible:
            return self._array.get_struct_initialization(variable_name, kernel_param_name, problem_id_substitute)
        return self.get_function_call_input(variable_name, kernel_param_name, problem_id_substitute, 'global')

    def get_kernel_parameters(self, kernel_param_name):
        return ['global {}* restrict {}'.format(self._ctype, kernel_param_name)]

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        if self._host_accessible:
            return self._array.enqueue_host_access(cl_environments, is_blocking=is_blocking, wait_for=wait_for)
        return {}

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        return {}

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        cl_context = cl_environment.context

        if self._host_accessible:
            return self._array.get_kernel_inputs(cl_environment, workgroup_size)

        mot_float_type_dtype = None
        if self._mot_float_dtype:
            mot_float_type_dtype = dtype_to_ctype(self._mot_float_dtype)
        dtype = np.dtype(ctype_to_dtype(self._ctype, mot_float_type_dtype))
        itemsize = dtype.itemsize

        if cl_context not in self._buffer_cache:
            if self._is_writable:
                if self._is_readable:
                    flags = cl.mem_flags.READ_WRITE
                else:
                    flags = cl.mem_flags.WRITE_ONLY
            else:
                flags = cl.mem_flags.READ_ONLY

            buffer = cl.Buffer(cl_context, flags, size=np.prod(self._shape) * itemsize)

            cl.enqueue_fill_buffer(cl_environment.queue, buffer,
                                   np.zeros(1, dtype=dtype), 0, np.prod(self._shape) * itemsize)

            self._buffer_cache[cl_context] = buffer

        return [self._buffer_cache[cl_context]]

    def get_nmr_kernel_inputs(self):
        return 1

    def _get_offset_str(self, problem_id_substitute):
        if self._parallelize_over_first_dimension:
            offset_str = str(self._data_length) + ' * {problem_id}'
        else:
            offset_str = '0'
        return offset_str.replace('{problem_id}', problem_id_substitute)


class CompositeArray(KernelData):

    def __init__(self, elements, ctype, address_space='private'):
        """An array filled with the given kernel data elements.

        Each of the given elements should be a :class:`Scalar` or an :class:`Array` with the property `as_scalar`
        set to True. We will load each value of the given elements into a private array.

        Args:
            elements (List[KernelData]): the kernel data elements to load into the private array
            ctype (str): the data type of this structure
            address_space (str): the address space for the allocation of the main array
        """
        self._elements = elements
        self._ctype = ctype
        self._address_space = address_space

        if self._address_space == 'private':
            self._composite_array = PrivateMemory(len(self._elements), self._ctype)
        elif self._address_space == 'local':
            self._composite_array = LocalMemory(self._ctype, len(self._elements))
        elif self._address_space == 'global':
            self._composite_array = Zeros(len(self._elements), self._ctype, mode='rw',
                                          parallelize_over_first_dimension=False)

    @property
    def ctype(self):
        return self._ctype

    def get_subset(self, problem_indices=None, batch_range=None):
        return self

    def set_mot_float_dtype(self, mot_float_dtype):
        for element in self._elements:
            element.set_mot_float_dtype(mot_float_dtype)
        self._composite_array.set_mot_float_dtype(mot_float_dtype)

    def get_data(self):
        return [item.get_data() for item in self._elements]

    def get_children(self):
        return self._elements

    def get_scalar_arg_dtypes(self):
        dtypes = list(self._composite_array.get_scalar_arg_dtypes())
        for d in self._elements:
            dtypes.extend(d.get_scalar_arg_dtypes())
        return dtypes

    def get_type_definitions(self):
        return '\n'.join(element.get_type_definitions() for element in self._elements)

    def initialize_variable(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return_str = self._composite_array.initialize_variable(variable_name, kernel_param_name,
                                                               problem_id_substitute, address_space)

        for ind, data in enumerate(self._elements):
            return_str += data.initialize_variable('{}_{}'.format(variable_name, str(ind)),
                                                   '{}_{}'.format(kernel_param_name, str(ind)),
                                                   problem_id_substitute, 'global')

            return_str += '{}[{}] = {};\n'.format(
                kernel_param_name,
                ind,
                data.get_struct_initialization(
                    '{}_{}'.format(variable_name, str(ind)),
                    '{}_{}'.format(kernel_param_name, str(ind)), problem_id_substitute))

        return return_str

    def get_function_call_input(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return self._composite_array.get_function_call_input(variable_name, kernel_param_name,
                                                             problem_id_substitute, address_space)

    def post_function_callback(self, variable_name, kernel_param_name, problem_id_substitute, address_space):
        return self._composite_array.post_function_callback(variable_name, kernel_param_name,
                                                            problem_id_substitute, address_space)

    def get_struct_declaration(self, name):
        return self._composite_array.get_struct_declaration(name)

    def get_struct_initialization(self, variable_name, kernel_param_name, problem_id_substitute):
        return self._composite_array.get_struct_initialization(variable_name, kernel_param_name, problem_id_substitute)

    def get_kernel_parameters(self, kernel_param_name):
        parameters = list(self._composite_array.get_kernel_parameters(kernel_param_name))
        for ind, d in enumerate(self._elements):
            parameters.extend(d.get_kernel_parameters('{}_{}'.format(kernel_param_name, str(ind))))
        return parameters

    def enqueue_host_access(self, cl_environments, is_blocking=True, wait_for=None):
        events = {}
        for d in self._elements:
            events.update(d.enqueue_host_access(cl_environments, is_blocking=is_blocking, wait_for=wait_for))
        return events

    def enqueue_device_access(self, cl_environments, is_blocking=True, wait_for=None):
        events = {}
        for d in self._elements:
            events.update(d.enqueue_device_access(cl_environments, is_blocking=is_blocking, wait_for=wait_for))
        return events

    def get_kernel_inputs(self, cl_environment, workgroup_size):
        data = list(self._composite_array.get_kernel_inputs(cl_environment, workgroup_size))
        for d in self._elements:
            data.extend(d.get_kernel_inputs(cl_environment, workgroup_size))
        return data

    def get_nmr_kernel_inputs(self):
        return self._composite_array.get_nmr_kernel_inputs() + \
               sum(element.get_nmr_kernel_inputs() for element in self._elements)
